use clap::Parser;
use ring::digest::Digest;

use std::io::Write;
use std::sync::atomic::{AtomicU64, Ordering};
use std::sync::mpsc::SyncSender;
use std::sync::{Arc, mpsc};
use std::time::Instant;

const REPORTING_QUANTUM: u64 = 1 << 16;

#[derive(Parser, Debug)]
struct Args {
    #[arg(short, long, default_value = "bitcoin is the future")]
    message: String,
    /// How many cpu threads to use
    #[arg(short, long)]
    threads: Option<usize>,
    /// Starting number
    #[arg(short, long, default_value = "0")]
    start: u64,
}

fn main() -> anyhow::Result<()> {
    let args = Args::parse();
    let threads = match args.threads {
        Some(t) => t,
        None => usize::from(std::thread::available_parallelism()?) / 2,
    };
    println!(
        "Running a microminer with \"{}\" base sentence starting with nonce {} in {} threads",
        args.message, args.start, threads
    );
    let (sender, receiver) = mpsc::sync_channel(128);
    let best_val = Arc::new(AtomicU64::new(u64::MAX >> 24)); // starting with 3 zero bytes
    for thread in 0..threads {
        let sender1 = sender.clone();
        let message = args.message.clone();
        let best_val_clone = best_val.clone();
        let start = args.start + thread as u64; // Incorporate args.start into thread start
        std::thread::spawn(move || {
            worker(
                message,
                start,
                threads as u64,
                best_val_clone,
                sender1,
            );
        });
    }
    reporting_task(receiver);

    Ok(())
}

fn reporting_task(receiver: mpsc::Receiver<Report>) {
    let mut last_log_time = Instant::now();
    let mut total_hash_count = 0_u64;
    let mut highest_nonce = 0_u64; // Track the highest nonce processed

    while let Ok(report) = receiver.recv() {
        match report {
            Report::Result(solution) => {
                let (hash, number) = hash_and_msb(solution.as_bytes());
                println!(
                    "hash(\"{solution}\") = {}, leading zero bits = {}",
                    hex::encode(hash),
                    number.leading_zeros()
                );
            }
            Report::PerThreadHashes(count) => {
                total_hash_count += count;
                log_hash_rate(&mut last_log_time, &mut total_hash_count, highest_nonce);
            }
            Report::Nonce(nonce) => {
                if nonce > highest_nonce {
                    highest_nonce = nonce; // Update to the highest nonce seen
                }
            }
        }
    }
}

fn log_hash_rate(last_log_time: &mut Instant, total_hashes: &mut u64, highest_nonce: u64) {
    if last_log_time.elapsed().as_secs() > 5 {
        let mhps = *total_hashes as f64 / (1000_000_f64 * last_log_time.elapsed().as_secs_f64());
        println!(
            "Average hash rate: {:.4} MH/s, Highest nonce processed: {}",
            mhps, highest_nonce
        );
        *last_log_time = Instant::now();
        *total_hashes = 0;
    }
}

enum Report {
    Result(String),
    PerThreadHashes(u64),
    Nonce(u64), // New variant to report the current nonce
}

fn worker(
    base_message: String,
    start: u64, // Now includes args.start + thread offset
    step: u64,
    best_val: Arc<AtomicU64>,
    reports: SyncSender<Report>,
) {
    let base_len = base_message.len();
    let mut raw_msg = base_message.into_bytes();

    for iter in 0..=u64::MAX {
        let nonce = start + (iter * step);
        write!(&mut raw_msg, "{nonce}").unwrap();
        let (_, hash_num) = hash_and_msb(&raw_msg);

        // Check zero bytes
        if hash_num < best_val.load(Ordering::Acquire) {
            best_val.store(hash_num, Ordering::Release);
            let result = String::from_utf8(raw_msg.clone()).unwrap();
            if reports.send(Report::Result(result)).is_err() {
                break;
            }
        }
        // Report periodically iterations done and current nonce
        if iter % REPORTING_QUANTUM == REPORTING_QUANTUM - 1 {
            if reports.send(Report::PerThreadHashes(REPORTING_QUANTUM)).is_err() {
                break;
            }
            if reports.send(Report::Nonce(nonce)).is_err() {
                break;
            }
        }
        raw_msg.truncate(base_len);
    }
}

/// Return hash and most significant 64 bits
fn hash_and_msb(data: &[u8]) -> (Digest, u64) {
    let first_round = ring::digest::digest(&ring::digest::SHA256, data);
    let hash2 = ring::digest::digest(&ring::digest::SHA256, first_round.as_ref());
    let hash_msb = <[u8; 8]>::try_from(&hash2.as_ref()[..8]).unwrap();
    let hash_num = u64::from_be_bytes(hash_msb);
    (hash2, hash_num)
}